import asyncio
import typing
import asyncio
from httpx._models import Request, Response
from httpx._transports.asgi import _ASGIApp
from httpx._transports.base import AsyncBaseTransport
from httpx._types import AsyncByteStream


class ASGIResponseByteStream(AsyncByteStream):
    def __init__(
        self, stream: typing.AsyncGenerator[bytes, None]
    ) -> None:
        self._stream = stream

    def __aiter__(self) -> typing.AsyncIterator[bytes]:
        return self._stream.__aiter__()

    async def aclose(self) -> None:
        await self._stream.aclose()


class ASGITransport(AsyncBaseTransport):
    def __init__(
        self,
        app: _ASGIApp,
        raise_app_exceptions: bool = True,
        root_path: str = "",
        client: typing.Tuple[str, int] = ("127.0.0.1", 123),
    ) -> None:
        self.app = app
        self.raise_app_exceptions = raise_app_exceptions
        self.root_path = root_path
        self.client = client

    async def handle_async_request(self, request: Request) -> Response:
        assert isinstance(request.stream, AsyncByteStream)

        # ASGI scope.
        scope = {
            "type": "http",
            "asgi": {"version": "3.0"},
            "http_version": "1.1",
            "method": request.method,
            "headers": [(k.lower(), v) for (k, v) in request.headers.raw],
            "scheme": request.url.scheme,
            "path": request.url.path,
            "raw_path": request.url.raw_path,
            "query_string": request.url.query,
            "server": (request.url.host, request.url.port),
            "client": self.client,
            "root_path": self.root_path,
        }

        # Request.
        request_body_chunks = request.stream.__aiter__()
        request_complete = False

        # Response.
        status_code = None
        response_headers = None
        sentinel = object()
        body_queue = asyncio.Queue()
        response_started = asyncio.Event()
        response_complete = asyncio.Event()

        # ASGI callables.

        async def receive() -> typing.Dict[str, typing.Any]:
            nonlocal request_complete

            if request_complete:
                await response_complete.wait()
                return {"type": "http.disconnect"}

            try:
                body = await request_body_chunks.__anext__()
            except StopAsyncIteration:
                request_complete = True
                return {"type": "http.request", "body": b"", "more_body": False}
            return {"type": "http.request", "body": body, "more_body": True}

        async def send(message: typing.Dict[str, typing.Any]) -> None:
            nonlocal status_code, response_headers, response_started, response_complete

            if message["type"] == "http.response.start":
                assert not response_started.is_set()
                status_code = message["status"]
                response_headers = message.get("headers", [])
                response_started.set()

            elif message["type"] == "http.response.body":
                assert response_started.is_set()
                assert not response_complete.is_set()
                body = message.get("body", b"")
                more_body = message.get("more_body", False)

                if body and request.method != "HEAD":
                    await body_queue.put(body)

                if not more_body:
                    await body_queue.put(sentinel)
                    response_complete.set()


        async def run_app() -> None:
            try:
                await self.app(scope, receive, send)
            except Exception:  # noqa: PIE-786
                if self.raise_app_exceptions or not response_complete.is_set():
                    raise

        async def body_stream() -> typing.AsyncGenerator[bytes, None]:
            while True:
                body = await body_queue.get()
                if body != sentinel:
                    yield body
                else:
                    return

        asyncio.create_task(run_app())

        await response_started.wait()
        assert status_code is not None
        assert response_headers is not None

        stream = ASGIResponseByteStream(body_stream())

        return Response(status_code, headers=response_headers, stream=stream)



from typing import Optional, Sequence, List
from ._proxy import create_handler
from fastapi.params import Depends


class router:
    """
    Unlock the full potential of Google Cloud Functions with FastAPI and async capabilities using the 'router' class!

    Are you looking to create a single powerful entry point for your Google Cloud Function?
    The 'router' class is designed for just that, providing seamless integration with FastAPI and async code.

    Key Highlights:
    - Easily register a single entry point for your Google Cloud Function.
    - Leverage FastAPI for robust request handling and request validation.
    - Harness the power of async code for responsive and scalable functions.
    - Define your function's input parameters with ease.

    Example Usage:

    ```python
    # Import necessary modules
    import asyncio
    from datetime import date
    from fastgcf import router

    # Simply use a decorator
    @router.get()
    async def main(start_date: date, end_date: date):
        await asyncio.sleep(1)  # Simulate async processing
        return {"start_date": start_date, "end_date": end_date}

    # That's it! Your Google Cloud Function is ready to handle and validate async GET requests seamlessly.
    ```
    """

    @staticmethod
    def get(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['GET'])

    @staticmethod
    def put(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['PUT'])

    @staticmethod
    def post(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['POST'])

    @staticmethod
    def options(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['OPTIONS'])

    @staticmethod
    def head(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['HEAD'])

    @staticmethod
    def patch(
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=['PATCH'])

    @staticmethod
    def route(
        dependencies: Optional[Sequence[Depends]] = None,
        methods: Optional[List[str]] = None,
    ):
        return create_handler(dependencies=dependencies, methods=methods)



import fastapi
import flask
import httpx
import asyncio
import functions_framework
import functools
import nest_asyncio
from fastapi.params import Depends
from fastapi.types import DecoratedCallable
from typing import Any, Callable, List, Optional, Sequence
from ._transport import ASGITransport


app = fastapi.FastAPI(docs_url=None, redoc_url=None, openapi_url=None)
is_entry_point_mounted = False
is_asyncio_patched = False
asgi_transport = ASGITransport(app)
async_client = httpx.AsyncClient(transport=asgi_transport, base_url='http://server')


def stream_response(response: httpx.Response, chunk_size: Optional[int] = None):
    chunk_size = chunk_size or 87040
    async_iterator = response.aiter_bytes(chunk_size=chunk_size)

    while True:
        try:
            yield asyncio.run(async_iterator.__anext__())
        except StopAsyncIteration:
            break


def convert_flask_request_to_httpx_request(request: flask.Request) -> httpx.Request:
    """
    Convert a Flask HTTP request to an HTTPX request.

    Args:
        request (flask.Request): The incoming Flask HTTP request.

    Returns:
        httpx.Request: An HTTPX request.
    """

    method = request.method
    url = request.url
    form_data = request.form.to_dict(flat=False) if len(request.form) > 0 else None
    headers = {key.lower(): value for key, value in request.headers}
    data = request.data
    params = request.args.to_dict(flat=False)
    cookies = {key: value for key, value in request.cookies.items()}

    files = []
    for file_name, file_stream in request.files.items(multi=True):
        file_headers = {key: value for key, value in file_stream.headers.items()}
        files.append((file_name, (file_stream.filename, file_stream.stream, file_stream.mimetype, file_headers)))

    has_files = len(files) > 0

    if form_data:
        data = form_data

    if len(data) == 0:
        data = None

    if not has_files:
        files = None

    if "content-length" in headers:
        del headers["content-length"]

    httpx_request = httpx.Request(
        method=method,
        url=url,
        headers=headers,
        data=data,
        params=params,
        cookies=cookies,
        files=files,
    )

    # Ensure that the request is encoded as 'application/x-www-form-urlencoded'
    # when form data is present and the 'files' object is falsy,
    # because HTTPX doesn't treat such requests as 'multipart/form-data'
    # even if the content type is set beforehand.
    if not has_files and form_data is not None:
        httpx_request.headers['content-type'] = 'application/x-www-form-urlencoded'

    return httpx_request


def convert_httpx_response_to_flask_response(response: httpx.Response) -> flask.Response:
    """
    Convert an HTTPX response to a Flask response.

    Args:
        response (httpx.Response): The HTTPX response.

    Returns:
        flask.Response: A Flask response.
    """

    flask_response = flask.Response(
        stream_response(response),
        status=response.status_code,
        content_type=response.headers.get('content-type')
    )

    for header, value in response.headers.items():
        flask_response.headers[header] = value

    for cookie in response.cookies.jar:
        flask_response.set_cookie(
            key=cookie.name,
            value=cookie.value,
            expires=cookie.expires,
            domain=cookie.domain,
            path=cookie.path,
            secure=cookie.secure
        )

    return flask_response


async def proxy(flask_request: flask.Request) -> flask.Response:
    """
    Proxy an incoming Flask request to the current FastAPI application.

    Args:
        flask_request (flask.Request): The incoming Flask request.

    Returns:
        flask.Response: The Flask response from the proxy operation.
    """

    httpx_request = convert_flask_request_to_httpx_request(flask_request)
    httpx_response = await async_client.send(httpx_request, stream=True)
    flask_response = convert_httpx_response_to_flask_response(httpx_response)
    return flask_response


def mount_entry_point(
    endpoint: Callable[..., Any],
    *,
    dependencies: Optional[Sequence[Depends]] = None,
    methods: Optional[List[str]] = None,
):
    """
    Mounts the entry point for the FastAPI application.

    This function registers the provided `endpoint` as the entry point for the FastAPI application.

    Args:
        endpoint (Callable[..., Any]): The function to be used as the entry point.
        dependencies (Optional[Sequence[Depends]], optional): A list of FastAPI dependencies.
            Defaults to None.
        methods (Optional[List[str]], optional): A list of HTTP methods that this entry point should respond to.
            Defaults to None.

    Raises:
        RuntimeError: If the entry point is already mounted.

    Note:
        This function should typically be called only once to set up the main entry point for the application.

    Example:
        mount_entry_point(my_endpoint_function, dependencies=[get_token])
    """

    global is_entry_point_mounted, is_asyncio_patched

    if is_entry_point_mounted:
        raise RuntimeError('Entry point is already mounted')

    if not is_asyncio_patched:
        nest_asyncio.apply()
        is_asyncio_patched = True

    app.add_api_route('/', endpoint, dependencies=dependencies, methods=methods)
    is_entry_point_mounted = True


def create_handler(
    dependencies: Optional[Sequence[Depends]] = None,
    methods: Optional[List[str]] = None,
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """
    Decorator for creating HTTP handlers that can be used as entry points.

    This decorator can be used to create HTTP handlers that will be treated as entry points
    in the Google Cloud Function acting like FastAPI handler.

    Args:
        dependencies (Optional[Sequence[Depends]], optional): A list of FastAPI dependencies.
            Defaults to None.
        methods (Optional[List[str]], optional): A list of HTTP methods that this handler should respond to.
            Defaults to None.

    Returns:
        Callable[[DecoratedCallable], DecoratedCallable]: A decorator function that can be applied to a handler function.
    """

    def decorator(func: DecoratedCallable) -> DecoratedCallable:
        mount_entry_point(func, dependencies=dependencies, methods=methods)
        func = functions_framework.http(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return asyncio.run(proxy(flask.request))

        return wrapper

    return decorator
